__pychecker__ = 'no-shadow'

import unittest
import time
import sys
import os

import rpc
from nfs4constants import *
from nfs4types import *
import nfs4lib
UID = -1
GID = -1

# Global variables
host = None
port = None
transport = "tcp"
prefix = None
sec_flavor = rpc.AUTH_UNIX
sec_extra = None  # Holds service type for RPCSEC_GSS

NCL1 = None
NCL2 = None
NCL3 = None
_cache_accepted = {}
_cache_rejected = {}

#
# nfs4st base classes
#

class SkipException(Exception):
    def __init__(self, msg=None):
        self.msg = msg

    def __str__(self):
        if self.msg:
            return self.msg
        else:
            return "Skipping test"
        
class MyTestCase(unittest.TestCase):
    def __init__(self, methodName='runTest'):
        """Create an instance of the class that will use the named test
           method when executed. Raises a ValueError if the instance does
           not have a method with the specified name.
        """
        unittest.TestCase.__init__(self,methodName)
        self.testMethod = getattr(self, methodName)
        self.verbose = TRUE

    def __call__(self, result=None):
        if result is None: result = self.defaultTestResult()
        else:
            self.verbose = result.showAll
        result.startTest(self)
        #testMethod = getattr(self, self.__testMethodName)
        try:
            try:
                self.setUp()
            except KeyboardInterrupt:
                raise
            except:
                result.addError(self, self._exc_info())
                return

            ok = 0
            try:
                self.testMethod()
                ok = 1
            except SkipException, e:
                result.addSkip(self, self._exc_info())
            except self.failureException, e:
                result.addFailure(self, self._exc_info())
            except KeyboardInterrupt:
                raise
            except:
                result.addError(self, self._exc_info())

            try:
                self.tearDown()
            except KeyboardInterrupt:
                raise
            except:
                result.addError(self, self._exc_info())
                ok = 0
            if ok: result.addSuccess(self)
        finally:
            result.stopTest(self)

    def _exc_info(self):
        """Return a version of sys.exc_info() with the traceback frame
           minimised; usually the top level of the traceback frame is not
           needed.
        """
        exctype, excvalue, tb = sys.exc_info()
        if sys.platform[:4] == 'java': ## tracebacks look different in Jython
            return (exctype, excvalue, tb)
        newtb = tb.tb_next
        if newtb is None:
            return (exctype, excvalue, tb)
        return (exctype, excvalue, newtb)

    def skipIf(self,expr,msg=None):
        if expr: raise SkipException, msg

    def skip(self,msg=None):
        raise SkipException, msg

class PartialTestClient:
    __pychecker__ = 'no-classattr'      # this class is a mix-in

    def __init__(self):
        # R/W recommended attributes with some garbage data
        self.__time = nfstime4(seconds=500000000, nseconds=0)
        self.writable_attrs = [
            ("acl",[nfsace4(ACE4_ACCESS_ALLOWED_ACE_TYPE,0,0,"EVERYONE@")]),
            ("archive",FALSE),
            ("hidden",FALSE),
            ("mimetype","nonsense"),
            ("mode",0644),
            ("owner","nonsense"),
            ("owner_group","nonsense"),
            ("system",FALSE),
            ("time_backup",self.__time),
            ("time_create",self.__time)
            ]
        # Readable recommended attributes with some garbage data
        self.readable_attrs = [
            ("acl",[nfsace4(ACE4_ACCESS_ALLOWED_ACE_TYPE,0,0,"EVERYONE@")]),
            ("aclsupport",0),
            ("archive",FALSE),
            ("cansettime",FALSE),
            ("case_insensitive",FALSE),
            ("case_preserving",FALSE),
            ("chown_restricted",FALSE),
            ("fileid",0x1234567890),
            ("files_avail",0x1234567890),
            ("files_free",0x1234567890),
            ("files_total",0x1234567890),
            #("fs_locations",FRED FIXME
            ("hidden",FALSE),
            ("homogeneous",FALSE),
            ("maxfilesize",0x1234567890),
            ("maxlink",0),
            ("maxname",0),
            ("maxread",0x1234567890),
            ("maxwrite",0x1234567890),
            ("mimetype","nonsense"),
            ("mode",0644),
            ("no_trunc",FALSE),
            ("numlinks",0),
            ("owner","nonsense"),
            ("owner_group","nonsense"),
            ("quota_avail_hard",0x1234567890),
            ("quota_avail_soft",0x1234567890),
            ("quota_used",0x1234567890),
            #("rawdev",FRED FIXME
            ("space_avail",0x1234567890),
            ("space_free",0x1234567890),
            ("space_total",0x1234567890),
            ("space_used",0x1234567890),
            ("system",FALSE),
            ("time_access",self.__time),
            ("time_backup",self.__time),
            ("time_create",self.__time),
            ("time_delta",self.__time),
            ("time_metadata",self.__time),
            ("time_modify",self.__time),
            ("mounted_on_fileid",0x1234567890)
            ]

    def lookup_all_objects(self):
        """Generate a list of lists with lookup operations for all types of objects"""
        return map(self.lookup_path, self.nfssuite.all_objects)

    def lookup_all_objects_and_sizes(self):
        return self.__lookup_objects_and_sizes(self.nfssuite.all_objects)

    def __lookup_objects_and_sizes(self, objects):
        """Generate a list of lists of lookup operations and object sizes"""
        obj_sizes = []
        for object in objects:
            lookupops = self.lookup_path(object)
            getattrop = self.getattr([FATTR4_SIZE])
            ops = [self.putrootfh_op()] + lookupops
            ops.append(getattrop)
            res = self.do_ops(ops)
            self.nfssuite.assert_OK(res)
            obj = res.resarray[-1].arm.arm.obj_attributes
            d =  nfs4lib.fattr2dict(obj)
            obj_sizes.append((lookupops, d[FATTR4_SIZE]))
        return obj_sizes

    def do_ops(self, *args, **kwargs):
        """Call compound. Handle all rpc.RPCExceptions as failures
        
        """
        return self.nfssuite.do_rpc(self.compound, *args, **kwargs)

    def create_confirm(self, attrs=None,mode=UNCHECKED4,
                       access=OPEN4_SHARE_ACCESS_BOTH, verifier=None,
                       error=[NFS4_OK], name=None, dir=None,
                       deny=OPEN4_SHARE_DENY_NONE, owner=None):
        """Create using open a regular file, and confirm the open

        Returns the fhandle and stateid from the confirm.
        """
        # Create the file
        if dir is None: dir=self.nfssuite.tmp_dir
        if name is None: name=self.nfssuite.obj_name
        if not attrs:
            attrs = nfs4lib.dict2fattr({FATTR4_MODE: 0644 })
        if (mode==EXCLUSIVE4) and (verifier==None):
            verifier = 0x12345678

        ops = [self.putrootfh_op()] + self.lookup_path(dir)
        ops.append(self.open(name, OPEN4_CREATE, mode, attrs, verifier,
                             access, deny, owner=owner))
        ops.append(self.getfh_op())
        res = self.do_ops(ops)
        self.nfssuite.assert_status(res, error)
        if res.status != NFS4_OK:
            return (None, None)
        fhandle = res.resarray[-1].arm.arm.object
        stateid = res.resarray[-2].arm.arm.stateid
        rflags = res.resarray[-2].arm.arm.rflags
        if not rflags & OPEN4_RESULT_CONFIRM:
            return (fhandle, stateid)

        # Confirm the open
        ops = [self.putfh_op(fhandle)]
        confirmop = self.open_confirm_op(stateid, self.get_seqid())
        ops.append(confirmop)
        res = self.do_ops(ops)
        self.nfssuite.assert_OK(res)
        stateid = res.resarray[-1].arm.arm.open_stateid
        return (fhandle, stateid)

    def open_confirm(self, dir=None, name=None, error=[NFS4_OK],
                     access=OPEN4_SHARE_ACCESS_READ, owner=None):
        """Open and confirm a noncreate open of an existing file
        """
        if not dir: dir=self.nfssuite.tmp_dir
        if not name: name=self.nfssuite.obj_name

        ops = [self.putrootfh_op()] + self.lookup_path(dir)
        ops.append(self.open(name, share_access=access, owner=owner))
        ops.append(self.getfh_op())
        res = self.do_ops(ops)
        self.nfssuite.assert_status(res, error)
        if res.status != NFS4_OK:
            return (None, None)
        fhandle = res.resarray[-1].arm.arm.object
        stateid = res.resarray[-2].arm.arm.stateid
        rflags = res.resarray[-2].arm.arm.rflags
        if not rflags & OPEN4_RESULT_CONFIRM:
            return (fhandle, stateid)

        # Confirm the open
        ops = [self.putfh_op(fhandle)]
        confirmop = self.open_confirm_op(stateid, self.get_seqid())
        ops.append(confirmop)
        res = self.do_ops(ops)
        self.nfssuite.assert_OK(res)
        stateid = res.resarray[-1].arm.arm.open_stateid

        return (fhandle, stateid)

    def open_claim_prev(self, fh, access=OPEN4_SHARE_ACCESS_READ,
                        deny=OPEN4_SHARE_DENY_NONE, seqid=None,
                        error=[NFS4_OK]):
        if seqid == None:
            seqid = self.get_seqid()
        ops = []
        if fh != None:
            ops.append(self.putfh_op(fh))
        claim = open_claim4(CLAIM_PREVIOUS,
                            delegate_type=OPEN_DELEGATE_NONE)
        owner = open_owner4(clientid=self.clientid,
                            owner=nfs4lib.long2opaque(self.get_owner()))
        how = openflag4(OPEN4_NOCREATE)
        ops.append(self.open_op(seqid, access, deny, owner, how, claim))
        ops.append(self.getfh_op())
        res = self.do_ops(ops)
        self.nfssuite.assert_status(res, error)
        if res.status != NFS4_OK:
            return (None, None)
        fhandle = res.resarray[-1].arm.arm.object
        stateid = res.resarray[-2].arm.arm.stateid
        return (fhandle, stateid)
        
    def close_file(self, fh, stateid, seqid=None, error=[NFS4_OK]):
        """close the given file
        """
        if not fh: return
        if seqid==None: seqid = self.get_seqid()
        
        ops = [self.putfh_op(fh)]
        ops.append(self.close_op(seqid, stateid))
        res = self.do_ops(ops)
        if error != None:
            self.nfssuite.assert_status(res, error)
    
    def read_file(self, offset=0, count=10, stateid=None, path=None,
              error=[NFS4_OK]):
        if not path:
            path = self.nfssuite.tmp_dir + [self.nfssuite.obj_name]
        if not stateid:
            stateid = stateid4(0, "")
        ops = [self.putrootfh_op()] + self.lookup_path(path)
        ops.append(self.read(stateid,offset,count))
        res = self.do_ops(ops)
        self.nfssuite.assert_status(res,error)
        if res.status == NFS4_OK:
            obj = res.resarray[-1].arm.arm
            return (obj.eof, obj.data)
    
    def write_file(self, offset=0, data="data", stateid=None, how=FILE_SYNC4,
               path=None, error=[NFS4_OK]):
        if not stateid:
            stateid = stateid4(0, "")
        if not path:
            path = self.nfssuite.tmp_dir + [self.nfssuite.obj_name]

        ops = [self.putrootfh_op()] + self.lookup_path(path)
        ops.append(self.write(data, stateid, offset, how))
        res = self.do_ops(ops)
        self.nfssuite.assert_status(res,error)
        if res.status == NFS4_OK:
            obj = res.resarray[-1].arm.arm
            return (obj.count, obj.committed)

    def lock_file(self, fh, stateid, offset=0, len=0xffffffffffffffff,
                 type=WRITE_LT, error=[NFS4_OK], owner="owner", seqid=None):
        """Lock the file in fh using owner for the first time

        Returns lock stateid, and the next lock_seqid should be 1
        """
        if seqid is None: seqid = self.get_seqid()
        ops = [self.putfh_op(fh)]
        lock_owner = lock_owner4(self.clientid, owner)
        open_owner = open_to_lock_owner4(seqid, stateid, 0, lock_owner)
        locker = locker4(TRUE, open_owner=open_owner)
        lockop = self.lock_op(type, FALSE, offset, len, locker)
        ops.append(lockop)
        res = self.do_ops(ops)
        self.nfssuite.assert_status(res,error)
        if res.status==NFS4_OK:
            return res.resarray[-1].arm.arm.lock_stateid

    def lock_test(self,fh,offset=0,len=50,type=WRITE_LT,error=[NFS4ERR_DENIED],
                 tester="tester"):
        """Test for a conflicting lock"""

        ops = [self.putfh_op(fh)]
        test_owner = lock_owner4(self.clientid, tester)
        lockop = self.lockt_op(type, offset, len, test_owner)
        ops.append(lockop)
        res = self.do_ops(ops)
        self.nfssuite.assert_status(res,error)
        return res

    def relock_file(self, fh, stateid, seqid, offset=0, len=0xffffffffffffffff,
               type=WRITE_LT, error=[NFS4_OK]):
        """Lock the file using stateid and seqid from previous lock operation
        """
        ops = [self.putfh_op(fh)]
        existing_lock_owner = exist_lock_owner4(stateid, seqid)
        locker = locker4(FALSE, lock_owner=existing_lock_owner)
        lockop = self.lock_op(type, FALSE, offset, len, locker)
        ops.append(lockop)
        res = self.do_ops(ops)
        self.nfssuite.assert_status(res,error)
        if res.status==NFS4_OK:
            return res.resarray[-1].arm.arm.lock_stateid

    def unlock_file(self, fh, stateid, seqid, offset=0, len=0xffffffffffffffff,
                    error=[NFS4_OK]):
        ops = [self.putfh_op(fh)]
        unlockop = self.locku_op(READ_LT, seqid, stateid, offset, len)
        ops.append(unlockop)
        res = self.do_ops(ops)
        self.nfssuite.assert_status(res,error)
        if res.status==NFS4_OK:
            return res.resarray[-1].arm.lock_stateid

class UDPTestClient(PartialTestClient, nfs4lib.UDPNFS4Client):
    def __init__(self, nfssuite, host, port=nfs4lib.NFS_PORT, uid=None, gid=None,name="", sec=None):
        self.nfssuite = nfssuite
        kwargs = {}
        if port: kwargs["port"] = port
        if uid != None: kwargs["uid"] = uid
        if gid != None: kwargs["gid"] = gid
        kwargs["name"] = name
        if sec != None: kwargs["sec"] = (sec_flavor, sec_extra)
        else:  kwargs["sec"] = sec
        nfs4lib.UDPNFS4Client.__init__(self, host, **kwargs)
        PartialTestClient.__init__(self)

class TCPTestClient(PartialTestClient, nfs4lib.TCPNFS4Client):
    def __init__(self, nfssuite, host, port=nfs4lib.NFS_PORT, uid=None, gid=None,name="", sec=None):
        self.nfssuite = nfssuite
        kwargs = {}
        if port: kwargs["port"] = port
        if uid != None: kwargs["uid"] = uid
        if gid != None: kwargs["gid"] = gid
        kwargs["name"] = name
        if sec == None: kwargs["sec"] = (sec_flavor, sec_extra)
        else:  kwargs["sec"] = sec
        nfs4lib.TCPNFS4Client.__init__(self, host, **kwargs)
        PartialTestClient.__init__(self)


class NFSSuite(MyTestCase):
    def __init__(self, methodName='runTest'):
        MyTestCase.__init__(self, methodName)
        self.obj_name = None

        # Filename constants. Same order as in nfs_ftype4 enum. 
        self.regfile = nfs4lib.unixpath2comps(prefix + "/doc/README") # NF4REG
        self.dirfile = nfs4lib.unixpath2comps(prefix + "/doc") # NF4DIR
        self.blockfile = nfs4lib.unixpath2comps(prefix + "/dev/fd0") # NF4BLK
        self.charfile = nfs4lib.unixpath2comps(prefix + "/dev/ttyS0") # NF4CHR
        self.linkfile = nfs4lib.unixpath2comps(prefix + "/dev/floppy") # NF4LNK
        self.socketfile = nfs4lib.unixpath2comps(prefix + "/dev/log") # NF4SOCK
        self.fifofile = nfs4lib.unixpath2comps(prefix + "/dev/initctl") # NF4FIFO

        # FIXME: Add NF4ATTRDIR and NF4NAMEDATTR types.
        self.special_objects = [self.blockfile, self.charfile, self.socketfile, self.fifofile]
        # Note named attribute dir or named attributes are not included in all_objects. 
        self.all_objects = self.special_objects + [self.regfile, self.dirfile, self.linkfile]

        # FIXME: Add sample named attribute.
        self.tmp_dir = nfs4lib.unixpath2comps(prefix + "/tmp")

        # Some more filenames
        self.hello_c = nfs4lib.unixpath2comps(prefix + "/src/hello.c")
        self.dirsymlinkfile = nfs4lib.unixpath2comps(prefix + "/src/doc")
        self.docporting = nfs4lib.unixpath2comps(prefix + "/doc/porting")
        self.tempfile = nfs4lib.unixpath2comps(prefix + "/tempfile")
        self.basedir = nfs4lib.unixpath2comps(prefix)
        
        # Should not exist
        self.vaporfilename = "vapor_object"
        self.vaporfile = nfs4lib.unixpath2comps(self.vaporfilename)
        # Not accessable
        self.notaccessibledir = nfs4lib.unixpath2comps(prefix + "/private")
        self.notaccessablefile = nfs4lib.unixpath2comps(prefix + "/private/info.txt")
        # Claim some ports for clients if this hasn't been sone yet
        self.connect()

    def create_client(self, uid, gid, name=""):
        if transport == "tcp":
            ncl = TCPTestClient(self, host, port, uid, gid, name=name)
        elif transport == "udp":
            ncl = UDPTestClient(self, host, port, uid, gid, name=name)
        else:
            raise RuntimeError, "Invalid protocol"
        return ncl

    def connect(self):
        global NCL1,NCL2,NCL3
        if not NCL1:
            if UID != -1:
                uid = UID
            elif os.getenv("PYNFS_UID"):
                uid = int(os.getenv("PYNFS_UID"))
            else:
                uid = 0
            if GID != -1:
                gid = GID
            elif os.getenv("PYNFS_GID"):
                gid = int(os.getenv("PYNFS_GID"))
            else:
                gid = 0
            # FRED - make uid/gid of ncl2 more sensible
            NCL1 = self.create_client(uid, gid, "client1")
            NCL2 = self.create_client(uid+1, gid, "client2")
            NCL3 = self.create_client(uid+2, gid, "client3")
            
        self.ncl  = NCL1
        self.ncl1 = NCL1
        self.ncl2 = NCL2
        self.ncl3 = NCL3

    def failIfRaises(self, excClass, callableObj, *args, **kwargs):
        """Fail if exception of excClass or EOFError is raised"""
        try:
            return apply(callableObj, args, kwargs)
        except excClass, e:
            self.fail(e)
        except EOFError, e:
            self.fail("EOFError: short response")
        return None

    def assert_OK(self, res):
        """Assert result from compound call is NFS4_OK"""
        self.assert_status(res, [NFS4_OK])

    def assert_status(self, res, errors):
        """Assert result from compound call is any of the values in errors"""
        if res.status in errors:
            return

        if res.resarray:
            lastop = res.resarray[-1]
            # res.status should be equal to lastop.arm.status
            if res.status != lastop.arm.status:
                e = nfs4lib.InvalidCompoundRes()
                self.fail(e)
            else:
                e = nfs4lib.BadCompoundRes(lastop.resop, res.status)
                self.fail(e)
        else:
            e = nfs4lib.EmptyBadCompoundRes(res.status)
            self.fail(e)

    def info_message(self, msg):
        #FRED - right now this prints before test header, should print after
        if self.verbose:
            print >> sys.stderr, "    " + msg + "\n",

    def do_rpc(self, method, *args, **kwargs):
        """Call method. Handle all rpc.RPCExceptions as failures"""
        return self.failIfRaises(rpc.RPCException, method, *args, **kwargs)

    def setUp(self):
        # Note: no network communication should be done in this method. 
        self.obj_name = "object1"
    
    def init_connection(self, *args, **kwargs):
        # FRED fixme - this might go in setup, need approp teardown
        ncl = kwargs.get("ncl")
        if ncl:
            del kwargs["ncl"]
        else:
            ncl = self.ncl
        self.failIfRaises(rpc.RPCException, ncl.init_connection,*args,**kwargs)

    def get_invalid_fh(self):
        """Return a (guessed) invalid filehandle"""
        return nfs4lib.long2opaque(123456780, NFS4_FHSIZE/8)

    def get_invalid_verifier(self):
        """Return a (guessed) invalid verifier"""
        return nfs4lib.long2opaque(123456780, NFS4_VERIFIER_SIZE/8)

    def get_invalid_clientid(self):
        """Return a (guessed) invalid clientid"""
        return 0x1234567890L

    def get_invalid_utf8strings(self):
        """Return a list of invalid ISO10646-UTF-8 strings"""
        # FIXME: More invalid strings. Do equivalence partitioning on ISO10646-UTF8 definition
        
        return ["\xc0\xc1", # starts two multibyte sequences
                "\xe0\x8a", # terminates a multibyte sequence too early
                "\xc0\xaf", # overlong character"
                "\xfc\x80\x80\x80\x80\xaf", # overlong character
                "\xfc\x80\x80\x80\x80\x80", # NULL
                "\xed\xa0\x80", # UTF-16 surrogate
                "\xed\xbf\xbf", # UTF-16 surrogate
                "\xef\xbf\xbe", # Invalid character U+FFFE
                "\xe3\xc0\xc0", # just mangled.
                "\xc0\x90", # overlong character
                # byte sequences that should never appear at start
                "\x80",
                "\xbf",
                "\xfe",
                "\xff",
                # starts with no ends
                "\xc0 ",
                "\xdf ",
                "\xe0 ",
                "\xef ",
                "\xf0 ",
                "\xf7 ",
                "\xf8 ",
                "\xfb ",
                "\xfc ",
                "\xfd "
                ]
                

    def remove_object(self, name=None, dir=None):
        """Remove object in dir, if it exists.

        Raises SkipException on failure
        """
        if not name: name = self.obj_name
        if not dir: dir = self.tmp_dir

        lookupops = self.ncl.lookup_path(dir)
        ops = [self.ncl.putrootfh_op()] + lookupops
        ops.append(self.ncl.remove_op(name))
        res = self.ncl.do_ops(ops)
        status = res.status in [NFS4_OK, NFS4ERR_NOENT]
        if not status:
            raise SkipException("Cannot remove %s from directory %s" % (name,dir))

    def create_obj(self, name=None, dir=None,type=NF4DIR):
        """Create non-file object 'name' in given directory.

        Raises SkipException on failure
        """
        if not name: name = self.obj_name
        if not dir: dir = self.tmp_dir

        lookupops = self.ncl.lookup_path(dir)
        ops = [self.ncl.putrootfh_op()] + lookupops
        
        if type in [NF4DIR,NF4SOCK,NF4FIFO]:
            objtype = createtype4(type=type)
        elif type == NF4LNK:
            objtype = createtype4(type=NF4LNK, linkdata="/etc/X11")
        elif type in [NF4BLK,NF4CHR]:
            devdata = specdata4(1, 2)
            objtype = createtype4(type=type, devdata=devdata)
        
        ops.append(self.ncl.create(objtype, name))
        res = self.ncl.do_ops(ops)
        if res.status != NFS4_OK:
            raise SkipException("Cannot create %s in directory %s" % (name,dir))

    def clean_dir(self, directory):
        """Clean directory. Raises SkipException on failure"""

        # Get fh, make sure is a dir
        lookup_dir_ops = self.ncl.lookup_path(directory)
        ops = [self.ncl.putrootfh_op()] + lookup_dir_ops
        attrs = nfs4lib.dict2fattr({FATTR4_TYPE:NF4DIR})
        ops.append(self.ncl.verify_op(attrs))
        ops.append(self.ncl.getfh_op())
        res = self.ncl.do_ops(ops)
        if res.status == NFS4ERR_NOT_SAME:
            raise SkipException("%s is not a directory" % directory)
        elif res.status != NFS4_OK:
            raise SkipException("Cannot clean directory %s" % directory)
        fh = res.resarray[-1].arm.arm.object

        # Read directory
        entries = self.do_rpc(self.ncl.do_readdir, fh)
        if not entries: return

        stateid = stateid4(0, "")
        attrs = nfs4lib.dict2fattr({FATTR4_MODE:0755})

        names = [entry.name for entry in entries]
        for name in names:
            # We separate setattr and remove to avoid an inode locking bug
            lookup_dir_ops = self.ncl.lookup_path(directory + [name])
            ops = [self.ncl.putrootfh_op()] + lookup_dir_ops
            ops.append(self.ncl.setattr_op(stateid, attrs))
            res = self.ncl.do_ops(ops)
            if res.status != NFS4_OK:
                raise SkipException("Cannot clean directory %s" % directory)

            lookup_dir_ops = self.ncl.lookup_path(directory)
            ops = [self.ncl.putrootfh_op()] + lookup_dir_ops
            ops.append(self.ncl.remove_op(name))
            res = self.ncl.do_ops(ops)

            #if res.status in (NFS4ERR_NOTEMPTY,NFS4ERR_EXIST):
            if res.status == NFS4ERR_NOTEMPTY:
                # Recursive deletion
                self.clean_dir(directory + [name])
                # Remove dir itself
                ops = [self.ncl.putrootfh_op()] + lookup_dir_ops
                ops.append(self.ncl.remove_op(name))
                res = self.ncl.do_ops(ops)
            if res.status != NFS4_OK:
                raise SkipException("Cannot clean directory %s" % directory)

        # Verify that all files were removed
        entries = self.do_rpc(self.ncl.do_readdir, fh)
        if entries:
            raise SkipException("Cannot clean directory %s" % directory)

    def make_sure_nonexistent(self, name=None, directory=None):
        """Make sure file does not exist. A READDIR is done. If file is not
        found in READDIR, return. If file is found in READDIR results, try to
        remove it.

        Returns true if file did not exist or it was possible to remove it.
        Returns false if file exists and it was not possible to remove it.
        """
        if not name:
            name = self.obj_name
        if not directory:
            directory = self.tmp_dir

        fh = self.do_rpc(self.ncl.do_getfh, directory)
        
        entries = self.do_rpc(self.ncl.do_readdir, fh)
        names = [entry.name for entry in entries]
        if name in names:
            # Strange, this file already exists
            # Try to remove it.
            try: self.remove_object(name, directory)
            except SkipException:
                return 0
        return 1
    
    def create_via_create(self, dstdir, filename):
        ops = [self.ncl.putrootfh_op()] + self.ncl.lookup_path(self.tmp_dir)
        objtype = createtype4(type=NF4DIR)
        ops.append(self.ncl.create(objtype, filename))
        return ops

    def create_via_link(self, dstdir, filename):
        operations = [self.ncl.putrootfh_op()]

        # Lookup source and save FH
        operations.extend(self.ncl.lookup_path(self.regfile))
        operations.append(self.ncl.savefh_op())

        # Lookup target directory
        operations.append(self.ncl.putrootfh_op())
        operations.extend(self.ncl.lookup_path(self.tmp_dir))

        operations.append(self.ncl.link_op(filename))
        return operations

    def create_via_open(self, dstdir=None, filename=None,
                        access=OPEN4_SHARE_ACCESS_BOTH):
        if not dstdir: dstdir = self.tmp_dir
        if not filename: filename = self.obj_name
        
        operations = [self.ncl.putrootfh_op()] + self.ncl.lookup_path(dstdir)
        attrs = nfs4lib.dict2fattr({FATTR4_MODE: 0644 })
        operations.append(self.ncl.open(filename,OPEN4_CREATE,UNCHECKED4,attrs,
                                        share_access=access))
        operations.append(self.ncl.getfh_op())
        return operations

    def create_via_rename(self, dstdir=None,filename=None):
        """ Because we call create_obj, this can raise SkipException"""
        if not dstdir: dstdir = self.tmp_dir
        if not filename: filename = self.obj_name

        # Create a directory, so we have something to rename
        oldname = "object1"
        try: self.create_obj(oldname)
        except SkipException, e:
            # if we raise again, let it fall through
            self.remove_object(oldname)
            self.create_obj(oldname)
        ops = [self.ncl.putrootfh_op()] + self.ncl.lookup_path(self.tmp_dir)
        ops.append(self.ncl.savefh_op())
        ops.append(self.ncl.putrootfh_op())
        ops.extend(self.ncl.lookup_path(self.tmp_dir))
        ops.append(self.ncl.rename_op(oldname, filename))

        return ops

    def accepted_filename(self, filename, remove_file=1, creator=None):
        """Returns true if file name could be created via creator.
        creator defaults to create_via_open. Files are created in /tmp
        """
        if not creator: creator = self.create_via_open
        
        ops = creator(self.tmp_dir, filename)
        
        res = self.ncl.do_ops(ops)
        # FIXME: Change to NFS4ERR_INVALIDNAME or what the error code will be called. 
        self.assert_status(res, [NFS4_OK, NFS4ERR_INVAL])
        creatable = (res.status == NFS4_OK)

        if creatable and remove_file:
            # Ok, the file could be created. 
            # Be nice and remove created object
            ops = [self.ncl.putrootfh_op()] + self.ncl.lookup_path(self.tmp_dir)
            ops.append(self.ncl.remove_op(filename))
            res = self.ncl.do_ops(ops)
            # Don't care about return status

        return creatable

    def try_file_names(self, remove_files=1, creator=None):
        """ Try all kinds of interesting file names and check if they are accepted
        (via accepted_filename). Creates files in self.tmp_dir. The optional argument
        remove_files can be unset to save created files. 

        Returns a tuple of (accepted_names, rejected_names)
        Raises SkipException if temporary directory cannot be cleared. You
        should catch it. 
        """
        if not creator:
            creator = self.create_via_open

        # Pulls 'creator' off of <bound method caller.creator ...>
        key = str(creator).split()[2].split('.')[-1]
        
        if _cache_accepted.has_key(key) and remove_files:
            return (_cache_accepted[key], _cache_rejected[key])
        
        try_names = self.get_invalid_utf8strings()
        rejected_names = []

        # aa<char>bb, with char 0-255
        for i in range(0, 256):
            ustr = "aa" + unichr(i) + "bb"
            try_names.append(ustr.encode("utf8"))

        # . and ..
        try_names.append(".".encode("utf8"))
        try_names.append("..".encode("utf8"))

        # C:
        try_names.append("C:".encode("utf8"))
        
        # Ok, lets try all these names.
        # Begin with cleaning /tmp
        self.clean_dir(self.tmp_dir)

        for filename in try_names[:]:
            # Do not remove the file after creation
            if not self.accepted_filename(filename, remove_file=remove_files,
                                           creator=creator):
                try_names.remove(filename)
                rejected_names.append(filename)

        _cache_accepted[key] = try_names
        _cache_rejected[key] = rejected_names
        return (try_names, rejected_names)

    def sleep(self, secs):
        """Sleep for given time, printing occasionally"""
        self.info_message("Sleeping for %d secs" % secs)
        time.sleep(secs)

    def supportedAttrs(self, file=None):
        """Returns bitmask of supported attributes"""
        if not file: file = self.tmp_dir + [self.obj_name]
        
        ops = [self.ncl.putrootfh_op()] + self.ncl.lookup_path(file)
        ops.append(self.ncl.getattr([FATTR4_SUPPORTED_ATTRS]))
        res = self.ncl.do_ops(ops)
        self.assert_OK(res)
        obj = res.resarray[-1].arm.arm.obj_attributes
        d = nfs4lib.fattr2dict(obj)
        return nfs4lib.intlist2long(d[FATTR4_SUPPORTED_ATTRS])

    def unsupportedAttrs(self, file=None):
        """ Return list of unsupported attributes for given file

        The list returns triples of the form (bitnumber,value,name)
        """
        # Get supported bitmask
        supp_bitmask = self.supportedAttrs(file)

        attrbitnum_dict = nfs4lib.get_attrbitnum_dict()
        unsupported = []
        for (name,value) in self.ncl.readable_attrs:
            bitnum = attrbitnum_dict[name]
            if not(supp_bitmask & (2**bitnum)):
                unsupported.append((bitnum,value,name))
        return unsupported

    def compareTimes(self, time1, time2):
        """Compares nfstime4 values

        Returns -1 if time1 < time2
                 0 if time1 ==time2
                 1 if time1 > time2
        """

        if time1.seconds < time2.seconds:
            return -1
        if time1.seconds > time2.seconds:
            return 1
        if time1.seconds == time2.seconds:
            if time1.nseconds < time2.nseconds:
                return -1
            if time1.nseconds > time2.nseconds:
                return 1
            return 0
         
    def makeBadID(self, stateid):
        """Given a good stateid, makes it bad

        NOTE this looks into server opaque data, thus is very specific
        to the CITI linux server
        """

        boottime = stateid.other[0:4]
        bad = stateid4(stateid.seqid , boottime+"\0\0\0\0\0\0\0\0")
        return bad

    def makeStaleID(self, stateid):
        """Given a good stateid, makes it stale

        NOTE this looks into server opaque data, thus is very specific
        to the CITI linux server
        """
        boottime = stateid.other[0:4]
        if ord(boottime[0]):
            staletime = "\0" + boottime[1:4]
        else:
            staletime = "a" + boottime[1:4]
        stale = stateid4(stateid.seqid , staletime+"\0\0\0\0\0\0\0\0")
        return stale

    def getLeaseTime(self):
        """Get length of lease time in seconds.

        If can't get lease time for some reason, defaults to 60
        """
        default = 60
        ops = [self.ncl.putrootfh_op()]
        ops.append(self.ncl.getattr([FATTR4_LEASE_TIME]))
        res = self.ncl.do_ops(ops)
        if res.status != NFS4_OK:
            return default
        obj = res.resarray[-1].arm.arm.obj_attributes
        d = nfs4lib.fattr2dict(obj)
        return d[FATTR4_LEASE_TIME]
