from base_st_classes import *

class AclSuite(NFSSuite):
    """Test POSIX style ACLs, probably quite CITI-centric

    This module is currently a work in progress
    """

    def setUp(self):
        NFSSuite.setUp(self)
        self.all = ACE4_READ_ATTRIBUTES | ACE4_READ_ACL # Must ALWAYS be set
        self.owner = ACE4_WRITE_ATTRIBUTES | ACE4_WRITE_ACL
        self.read_mode = ACE4_READ_DATA | ACE4_READ_NAMED_ATTRS | self.all
        self.write_mode = ACE4_WRITE_DATA | ACE4_WRITE_NAMED_ATTRS | ACE4_APPEND_DATA | ACE4_DELETE_CHILD | self.all 
        self.execute_mode = ACE4_EXECUTE | self.all
        self.deny = ~0

    def tearDown(self):
        try: self.clean_dir(self.tmp_dir)
        except SkipException, e:
            pass

    def _addUsers(self,acelist,stdmode,userlist=[],mask=0):
        """ Add user/mode pairs in userlist to acelist

        Permissions will be restricted to only those allowed in mask.
        A named user must have 3 consecutive entries:
        DENY  flag   mask   who
        ALLOW  ""   access  ""
        DENY   ""  ~access  ""
        """
        
        ace = nfsace4(ACE4_ACCESS_ALLOWED_ACE_TYPE,0,stdmode|self.owner,"OWNER@")
        acelist.append(ace)
        ace = nfsace4(ACE4_ACCESS_DENIED_ACE_TYPE,0,self.deny&~(stdmode|self.owner),"OWNER@")
        acelist.append(ace)
        
        for (user,mode) in userlist:
            ace = nfsace4(ACE4_ACCESS_DENIED_ACE_TYPE,0,self.deny&~mask,user)
            acelist.append(ace)
            ace = nfsace4(ACE4_ACCESS_ALLOWED_ACE_TYPE,0,mode,user)
            acelist.append(ace)
            ace = nfsace4(ACE4_ACCESS_DENIED_ACE_TYPE,0,self.deny&~mode,user)
            acelist.append(ace)

    def _addGroups(self,acelist,stdmode,grouplist=[],mask=0):
        """ Add group/mode pairs in grouplist to acelist

        Permissions will be restricted to only those allowed in mask.
        A group must have 3 entries:
        DENY  flag   mask   who
        ALLOW  ""   access  ""
        DENY   ""  ~access  ""
        For groups, the first two lines of every group must proceed any groups
        third line.
        """
        # Start std group access
        ace = nfsace4(ACE4_ACCESS_DENIED_ACE_TYPE,0,self.deny&~mask,"GROUP@")
        acelist.append(ace)
        ace = nfsace4(ACE4_ACCESS_ALLOWED_ACE_TYPE,0,stdmode,"GROUP@")
        acelist.append(ace)

        # Start named groups
        for (group,mode) in grouplist:
            ace = nfsace4(ACE4_ACCESS_DENIED_ACE_TYPE,
                          ACE4_IDENTIFIER_GROUP,self.deny&~mask,group)
            acelist.append(ace)
            ace = nfsace4(ACE4_ACCESS_ALLOWED_ACE_TYPE,
                          ACE4_IDENTIFIER_GROUP,mode,group)
            acelist.append(ace)

        # Finish std group
        ace = nfsace4(ACE4_ACCESS_DENIED_ACE_TYPE,0,self.deny&~stdmode,"GROUP@")
        acelist.append(ace)
        
        # Finish named groups
        for (group,mode) in grouplist:
            ace = nfsace4(ACE4_ACCESS_DENIED_ACE_TYPE,
                          ACE4_IDENTIFIER_GROUP,self.deny&~mode,group)
            acelist.append(ace)

    def _addOther(self,acelist,mode):
        ace = nfsace4(ACE4_ACCESS_ALLOWED_ACE_TYPE,0,mode,"EVERYONE@")
        acelist.append(ace)
        ace = nfsace4(ACE4_ACCESS_DENIED_ACE_TYPE,0,self.deny&~mode,"EVERYONE@")
        acelist.append(ace)

    def _setACL(self, user, group=0, other=0, mask=None,
                userlist = [], grouplist=[],
                error=[NFS4_OK]):
        if not mask: mask = self.read_mode
        if not group: group = self.all
        if not other: other = self.all
        
        file = self.tmp_dir + [self.obj_name]
        stateid = stateid4(0, "")
        ops = [self.ncl.putrootfh_op()] + self.ncl.lookup_path(file)
        acelist = []
        self._addUsers(acelist, user, userlist, mask)
        self._addGroups(acelist, group, grouplist, mask)
        self._addOther(acelist, other)
        attrs = nfs4lib.dict2fattr({FATTR4_ACL: acelist})
        ops.append(self.ncl.setattr_op(stateid, attrs))
        res = self.ncl.do_ops(ops)
        self.assert_status(res,error)
        if res.status == NFS4_OK:
            return res
    
    def _checkMode(self,mode,mask=None):
        """ Assert that mode attr of self.obj_name equals 'mode'

        This is complicated by the fact that the mask is handled strangely.
        If a mask exists, ITS mode is returned for group mode.

        So if Mask==None, we don't compare group at all.
        Otherwise mask in [0..7] and we set mode compare accordingly.

        Further note...above is POSIX behavior, but shouldn't the server be
        following the RFC, which seems to state that the group mode bits
        should correspond to the GROUP@ entries, not the mask 
        """
        file = self.tmp_dir + [self.obj_name]
        ops = [self.ncl.putrootfh_op()] + self.ncl.lookup_path(file)
        ops.append(self.ncl.getattr([FATTR4_MODE]))
        res = self.ncl.do_ops(ops)
        self.assert_OK(res)
        obj = res.resarray[-1].arm.arm.obj_attributes
        d = nfs4lib.fattr2dict(obj)
        if mask != None:
            mode = (mode&07707) | ((mask&7)*8)
            self.assertEqual(mode,d[FATTR4_MODE],"Returned mode %o does not equal expected mode %o"%(d[FATTR4_MODE],mode))
        else:
            self.assertEqual(mode&07707,d[FATTR4_MODE]&07707,"Returned mode %o does not equal expected mode %o"%(d[FATTR4_MODE],mode))
            
    #
    # TESTS
    #
    def testGetAllObjects(self):
        """Try to get ACL on all type of objects
        
        """
        for lookupops in self.ncl.lookup_all_objects():
            ops = [self.ncl.putrootfh_op()] + lookupops
            ops.append(self.ncl.getattr([FATTR4_ACL]))
            res = self.ncl.do_ops(ops)
            self.assert_OK(res)

            #obj = res.resarray[-1].arm.arm.obj_attributes
            #d = nfs4lib.fattr2dict(obj)
            #print d

    def testSetFile(self):
        """Try to set an ACL on a file
        """
        # Create a file with mode=0644
        self.init_connection()
        ops = self.create_via_open()
        res = self.ncl.do_ops(ops)
        self.assert_OK(res)

        # Set ACL with mode=0700
        mode = self.read_mode | self.write_mode | self.execute_mode
        self._setACL(mode)

        # Check mode was reset
        self._checkMode(0700)
        
    def testObjects(self):
        """Try to set ACL objects
        """
        self.init_connection()
        for type in [ NF4DIR,
                      NF4BLK,
                      NF4CHR,
                      # NF4LNK, # Tested seperately
                      NF4SOCK,
                      NF4FIFO ]:
            try: self.create_obj(type = type)
            except SkipException, e:
                self.fail(e)

            # Set ACL with mode=0700
            mode = self.read_mode | self.write_mode | self.execute_mode
            self._setACL(mode)
            
            # Check mode was reset
            self._checkMode(0700)

            # Remove created object
            ops = [self.ncl.putrootfh_op()] + self.ncl.lookup_path(self.tmp_dir)
            ops.append(self.ncl.remove_op(self.obj_name))
            res = self.ncl.do_ops(ops)
            self.assert_OK(res)

    def testSymlink(self):
        """Try to set ACL on symlink"""

        try: self.create_obj(type = NF4LNK)
        except SkipException, e:
            self.skip(e)
        
        # Set ACL with mode=0700
        mode = self.read_mode | self.write_mode | self.execute_mode
        self._setACL(mode)
        
        # Check mode was reset
        self._checkMode(0700)

    def testEmptyACL(self):
        """Try to send an empty ACL"""
        try: self.create_obj()
        except SkipException, e:
            self.skip(e)

        file = self.tmp_dir + [self.obj_name]
        ops = [self.ncl.putrootfh_op()] + self.ncl.lookup_path(file)
        stateid = stateid4(0, "")
        attrs = nfs4lib.dict2fattr({FATTR4_ACL: []})
        ops.append(self.ncl.setattr_op(stateid, attrs))
        res = self.ncl.do_ops(ops)
        self.assert_status(res, [NFS4_OK,NFS4ERR_INVAL])

    def testShortAcl(self):
        """Try to send a length one ACL"""
        try: self.create_obj()
        except SkipException, e:
            self.skip(e)

        file = self.tmp_dir + [self.obj_name]
        ops = [self.ncl.putrootfh_op()] + self.ncl.lookup_path(file)
        stateid = stateid4(0, "")
        acelist = [nfsace4(ACE4_ACCESS_DENIED_ACE_TYPE,0,self.deny&~self.all,"EVERYONE@")]
        #acelist = []
        #self._addUsers(acelist, self.read_mode)
        #self._addGroups(acelist, self.all)
        #self._addOther(acelist, self.all)
        attrs = nfs4lib.dict2fattr({FATTR4_ACL: acelist})
        ops.append(self.ncl.setattr_op(stateid, attrs))
        res = self.ncl.do_ops(ops)
        self.assert_status(res, [NFS4ERR_INVAL])

    def testLongACL(self):
        """Try to set ACL, with lots of entries"""
        self.info_message("Need general way of getting valid user/group names")
        
        # Create a file with mode=0644
        self.init_connection()
        ops = self.create_via_open()
        res = self.ncl.do_ops(ops)
        self.assert_OK(res)

        # Give different users access
        userlist = [ ("adm@citi.umich.edu",self.read_mode),
                     ("fisaman@citi.umich.edu",self.read_mode),
                     ("iisaman@citi.umich.edu",self.read_mode)
                     ]

        # Give groups access
        grouplist = [ ("root@citi.umich.edu",self.read_mode),
                      ("bin@citi.umich.edu",self.read_mode),
                      ("daemon@citi.umich.edu",self.read_mode),
                      ("sys@citi.umich.edu",self.read_mode),
                      ("adm@citi.umich.edu",self.read_mode),
                      ("tty@citi.umich.edu",self.read_mode),
                      ("disk@citi.umich.edu",self.read_mode),
                      ("lp@citi.umich.edu",self.read_mode),
                      ("mem@citi.umich.edu",self.read_mode),
                      ("kmem@citi.umich.edu",self.read_mode),
                      ("wheel@citi.umich.edu",self.read_mode),
                      ("mail@citi.umich.edu",self.read_mode),
                      ("news@citi.umich.edu",self.read_mode),
                      ("uucp@citi.umich.edu",self.read_mode),
                      ("man@citi.umich.edu",self.read_mode),
                      ("games@citi.umich.edu",self.read_mode),
                      ("gopher@citi.umich.edu",self.read_mode),
                      ("dip@citi.umich.edu",self.read_mode),
                      ("ftp@citi.umich.edu",self.read_mode),
                      ("lock@citi.umich.edu",self.read_mode),
                      ("nobody@citi.umich.edu",self.read_mode),
                      ("users@citi.umich.edu",self.read_mode),
                      ("rpm@citi.umich.edu",self.read_mode),
                      ("floppy@citi.umich.edu",self.read_mode),
                      ("vcsa@citi.umich.edu",self.read_mode),
                      ("utmp@citi.umich.edu",self.read_mode),
                      ("nscd@citi.umich.edu",self.read_mode),
                      ("slocate@citi.umich.edu",self.read_mode),
                      ("sshd@citi.umich.edu",self.read_mode),
                      ("rpc@citi.umich.edu",self.read_mode),
                      ("mailnull@citi.umich.edu",self.read_mode),
                      ("smmsp@citi.umich.edu",self.read_mode),
                      ("pcap@citi.umich.edu",self.read_mode),
                      ("xfs@citi.umich.edu",self.read_mode),
                      ("ntp@citi.umich.edu",self.read_mode),
                      ("gdm@citi.umich.edu",self.read_mode),
                      ("desktop@citi.umich.edu",self.read_mode),
                      ("fisaman@citi.umich.edu",self.read_mode),
                      ("rpcuser@citi.umich.edu",self.read_mode)
                      ]

        # Set ACL with mode=0700
        mode = self.read_mode | self.write_mode | self.execute_mode
        self._setACL(mode, userlist=userlist, grouplist=grouplist,
                     error=[NFS4ERR_INVAL])
        
    def testNamedACL(self):
        """Try to set ACL, with named entries"""

        self.info_message("Need general way of getting valid user/group names")
        
        # Create a file with mode=0644
        self.init_connection()
        ops = self.create_via_open()
        res = self.ncl.do_ops(ops)
        self.assert_OK(res)

        # Give different users access
        userlist = [ ("root@citi.umich.edu",self.read_mode),
                     ("bin@citi.umich.edu",self.read_mode),
                     ("adm@citi.umich.edu",self.read_mode)
                     ]

        # Give groups access
        grouplist = [ ("root@citi.umich.edu",self.read_mode),
                      ("bin@citi.umich.edu",self.read_mode),
                      ("adm@citi.umich.edu",self.read_mode)
                      ]

        # Set ACL with mode=0700
        mode = self.read_mode | self.write_mode | self.execute_mode
        self._setACL(mode, userlist=userlist, grouplist=grouplist)
        
    def xtestBadlyNamedACL(self):
        """Try to set ACL, with incorrectly named entries"""
        # FRED - This return OK.  If this is correct behavior, we can
        # use to generate large lists.  (NOPE, it appears first nonsense
        # one is mapped to nobody, and a second (different) one causes return of
        # invalid because of duplication
        
        # Create a file with mode=0644
        self.init_connection()
        ops = self.create_via_open()
        res = self.ncl.do_ops(ops)
        self.assert_OK(res)

        # Give different users access
        userlist = [ ("gobdlygook",self.read_mode) ]
        
        # Give groups access
        grouplist = [ ("moregobdlygook@gibberish.org",self.read_mode) ]
        
        # Set ACL with mode=0700
        mode = self.read_mode | self.write_mode | self.execute_mode
        self._setACL(mode, userlist=userlist, grouplist=grouplist,
                     error=[NFS4ERR_INVAL])

    def testBadOrder(self):
        """Try ACL with entries in wrong order"""
        self.init_connection()
        ops = self.create_via_open()
        res = self.ncl.do_ops(ops)
        self.assert_OK(res)
        
        mask = self.read_mode
        group = "root@citi.umich.edu"
        mode = self.read_mode
        
        file = self.tmp_dir + [self.obj_name]
        stateid = stateid4(0, "")
        ops = [self.ncl.putrootfh_op()] + self.ncl.lookup_path(file)
        acelist = []
        self._addUsers(acelist, self.read_mode)

        # Wrong group order
        '''
        ace = nfsace4(ACE4_ACCESS_DENIED_ACE_TYPE,0,self.deny&~mask,"GROUP@")
        acelist.append(ace)
        ace = nfsace4(ACE4_ACCESS_DENIED_ACE_TYPE,
                      ACE4_IDENTIFIER_GROUP,self.deny&~mask,group)
        acelist.append(ace)
        ace = nfsace4(ACE4_ACCESS_ALLOWED_ACE_TYPE,
                      ACE4_IDENTIFIER_GROUP,mode,group)
        acelist.append(ace)
        ace = nfsace4(ACE4_ACCESS_DENIED_ACE_TYPE,
                      ACE4_IDENTIFIER_GROUP,self.deny&~mode,group)
        acelist.append(ace)
        '''

        ace = nfsace4(ACE4_ACCESS_ALLOWED_ACE_TYPE,0,0,"GROUP@")
        acelist.append(ace)
        ace = nfsace4(ACE4_ACCESS_DENIED_ACE_TYPE,0,self.deny,"GROUP@")
        acelist.append(ace)
        ace = nfsace4(ACE4_ACCESS_DENIED_ACE_TYPE,0,self.deny&~mask,"GROUP@")
        acelist.append(ace)

        # End of group stuff
        self._addOther(acelist, 0)
        attrs = nfs4lib.dict2fattr({FATTR4_ACL: acelist})
        ops.append(self.ncl.setattr_op(stateid, attrs))
        res = self.ncl.do_ops(ops)
        self.assert_status(res, [NFS4ERR_INVAL])
