#!/usr/bin/env python2
# -*- coding: utf-8 -*-

# nfs4st.py - NFS4 server tester
#
# Written by Peter strand <peter@cendio.se>
# Copyright (C) 2001 Cendio Systems AB (http://www.cendio.se)
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License. 
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

# TODO:
#
# Extend unittest with warnings.
#
# Handle errors such as NFS4ERR_RESOURCE and NFS4ERR_DELAY.
#
# filehandles are split into eq. classes "valid filehandle" and
# "invalid filehandle". There should probably be a class "no filehandle" as
# well. Currently, "invalid filehandle" are tested by doing operations without
# filehandles.
#
# Add testing of \ to testSlash methods.
#
# More testing of strange attributes and handling of NFS4ERR_ATTRNOTSUPP;
# more fine-grained eqv.part of attribute masks. 
#
# Nomenclature: Each test class is referred to as a "test suite". Each
# test* method is a "test case".

import nfs4st
from nfs4st import *
from nfs4st.base_st_classes import *
import types

#
# nfs4st test execution code
#


class MyTextTestResult(unittest._TextTestResult):
    def __init__(self, print_tracebacks, stream, descriptions, verbosity):
        unittest._TextTestResult.__init__(self, stream, descriptions, verbosity)
        self.print_tracebacks = print_tracebacks
        self.skips = []
        print
    def addSkip(self, test, err):
        """Called when a test is skipped. 'err' is a tuple of values as
        returned by sys.exc_info().
        """
        self.skips.append((test, self._exc_info_to_string(err)))
        if self.showAll:
            self.stream.writeln("skip")
        elif self.dots:
            self.stream.write('S')

    def printErrors(self):
        if self.print_tracebacks > 1:
            self.printErrorList('SKIP', self.skips)
        if self.print_tracebacks > 0:
            self.printErrorList('FAIL', self.failures)
        self.printErrorList('ERROR', self.errors)


class MyTextTestRunner:
    def __init__(self, verbosity=1, print_tracebacks=1, stream=sys.stdout, descriptions=1):
        self.stream = unittest._WritelnDecorator(stream)
        self.descriptions = descriptions
        # 0=no output, 1=dots, 2=test description and info messages
        self.verbosity = verbosity
        self.print_tracebacks = print_tracebacks
    
    def _makeResult(self):
        return MyTextTestResult(self.print_tracebacks, self.stream, self.descriptions, self.verbosity)

    def run(self,test):
        "Run the given test case or test suite."
        result = self._makeResult()
        startTime = time.time()
        test(result)
        stopTime = time.time()
        timeTaken = float(stopTime - startTime)
        result.printErrors()
        self.stream.writeln()
        self.stream.writeln(result.separator2)
        run = result.testsRun
        self.stream.writeln("Ran %d test%s in %.3fs" %
                            (run, run == 1 and "" or "s", timeTaken))
        self.stream.writeln()
        if not result.wasSuccessful() or result.skips:
            if result.wasSuccessful():
                self.stream.write("PASSED (")
            else:
                self.stream.write("FAILED (")
            skipped, failed, errored = map(len, (result.skips, result.failures, result.errors))
            if skipped:
                self.stream.write("skipped=%d" % skipped)
            if failed:
                if skipped: self.stream.write(", ")
                self.stream.write("failures=%d" % failed)
            if errored:
                if failed or skipped: self.stream.write(", ")
                self.stream.write("errors=%d" % errored)
            self.stream.writeln(")")
        else:
            self.stream.writeln("OK")
        return result
        
class MyTestLoader(unittest.TestLoader):
    def loadTestsFromName(self, name, module=None):
        """Return a suite of all tests cases given a string specifier.

        The name may resolve either to a module, a test case class, a
        test method within a test case class, or a callable object which
        returns a TestCase or TestSuite instance.

        The method optionally resolves the names relative to a given module.
        """
        parts = name.split('.')
        if module is None:
            if not parts:
                raise ValueError, "incomplete test name: %s" % name
            else:
                parts_copy = parts[:]
                while parts_copy:
                    try:
                        module = __import__('.'.join(parts_copy))
                        break
                    except ImportError:
                        del parts_copy[-1]
                        if not parts_copy: raise
                parts = parts[1:]
        obj = module
        old_part = None
        for part in parts:
            try:
                obj = getattr(obj, part)
            except AttributeError, e:
                if old_part and old_part.startswith("st_"):
                    suite_name = old_part[3:].capitalize() + "Suite"
                else:
                    raise
                obj = getattr(obj, suite_name)
                obj = getattr(obj, part)
            old_part = part

        if type(obj) == types.ModuleType:
            return self.loadTestsFromModule(obj)
        elif type(obj) == types.ClassType and issubclass(obj, unittest.TestCase):
            return self.loadTestsFromTestCase(obj)
        elif type(obj) == types.UnboundMethodType:
            return obj.im_class(obj.__name__)
        elif callable(obj):
            test = obj()
            if not isinstance(test, unittest.TestCase) and \
               not isinstance(test, unittest.TestSuite):
                raise ValueError, \
                      "calling %s returned %s, not a test" % (obj,test)
            return test
        else:
            raise ValueError, "don't know how to make test from: %s" % obj


class TestProgram(unittest.TestProgram):
    USAGE = """\
Usage: %(progName)s [nfs://]host[:port]<prefix> [options] [test] [...]

<prefix> defaults to /. Use same prefix as for test_tree_net.py 

Options:
  -u, --udp            use UDP as transport 
  -t, --tcp            use TCP as transport (default)
  -h, --help           Show this message
  -q                   Minimal output during testing
  -Q, --quiet          No ouput during testing
  -v                   display failure tracebacks
  -V, --verbose        display all tracebacks
      --uid=<uid>      Use custom <uid> with AUTH_UNIX security
      --gid=<gid>      Use custom <gid> with AUTH_UNIX security
      --owner=         Use custom fattr_owner in relevant requests.
      --owner_group=   Use custom fattr_owner in relevant requests.
      --sec=           Use given security (default=sys)

Examples:
  %(progName)s                        - run default set of tests
  %(progName)s st_create              - run tests from 'st_create.py'
"""
    def parseArgs(self, argv):
        import getopt
        import re
        
        self.verbosity = 2
        self.print_tracebacks = 0
        self.testNames = []

        # Reorder arguments, so we can add options at the end 
        options = []
        args = []
        argsleft = argv[1:]
        scan_done = 0
        while not scan_done:
            try:
                scan_opt, scan_arg = getopt.getopt(argsleft, 'uthqQvV',
                                   ['help', 'quiet', 'udp', 'tcp', 'verbose',
                                    'uid=', 'gid=', 'owner=', 'owner_group=',
                                    'sec='
                                    ])
            except getopt.error, msg:
                self.usageExit(msg)
            options += scan_opt
            if len(scan_arg) < 1:
                scan_done = 1
            else:
                args += [scan_arg[0]]
                argsleft = scan_arg[1:]

        for opt, value in options:
            if opt == "--sec":
                import rpc
                try:
                    flavor, extra = rpc.supported[value]
                except KeyError:
                    print "Unsupported option for --sec : %s" % value
                    print "Supported options are:"
                    print rpc.supported.keys()
                    sys.exit(2)
                base_st_classes.sec_flavor = flavor
                base_st_classes.sec_extra = extra
            if opt == "--uid":
                base_st_classes.UID = int(value)
            if opt == "--gid":
                base_st_classes.GID = int(value)
            if opt == "--owner":
                nfs4lib.OWNER = str(value)
            if opt == "--owner_group":
                nfs4lib.OWNER_GROUP = str(value)
            if opt in ("-u", "--udp"):
                base_st_classes.transport = "udp"
            if opt in ("-t", "--tcp"):
                base_st_classes.transport = "tcp"
            if opt in ("-h","--help"):
                self.usageExit()
            if opt == "-q":
                self.verbosity = 1
            if opt in ("-Q","--quiet"):
                self.verbosity = 0
            if opt in ("-V","--verbose"):
                self.print_tracebacks = 2
            if opt == "-v":
                self.print_tracebacks = 1

        if len(args) < 1:
            self.usageExit()

        parse_result = nfs4lib.parse_nfs_url(args[0])
        if not parse_result:
            self.usageExit()

        (nfs4st.base_st_classes.host, portstring, directory) = parse_result

        if not directory:
            directory = "/"
            
        base_st_classes.prefix = directory + "/nfs4st_tree"

        if portstring:
            base_st_classes.port = int(portstring)
        else:
            base_st_classes.port = nfs4lib.NFS_PORT

        args = args[1:]

        # Runs all test suites if no arguments are passed.
        if len(args) == 0 and self.defaultTest is None:
            global_list = globals().keys()
            for suite in global_list:
                if re.match("st_", suite) != None:
                    self.testNames.append(suite)

        elif len(args) > 0:
            self.testNames = args
        else:
            self.testNames = (self.defaultTest,)

        self.createTests()

    def runTests(self):
        self.testRunner = MyTextTestRunner(self.verbosity,self.print_tracebacks)
        result = self.testRunner.run(self.test)
        sys.exit(not result.wasSuccessful())


main = TestProgram(testLoader=MyTestLoader())

if __name__ == "__main__":
    main()
